import dimod
import dwave as dw
import minorminer
import networkx as nx
import numpy as np
import statistics as stat

import dwave.preprocessing.lower_bounds as dwplb
import dwave.embedding as dwe

# Makes the h and J np arrays for use in creating the bqm and networkx graph
def make_h_J(C_i, C_ij, mu, sigma):
    h = 2 * sigma * (np.einsum('ij, j', C_ij, mu) - C_i)
    J = 2 * np.triu(C_ij, k=1) * pow(sigma, 2)
    
    return h, J

# Independent function for simplifying problem
# -> should hold a couple simple pruning methods
def default_prune(J, cutoff_percentile):
    rows, cols = np.shape(J)
    sign_J = np.ndarray.flatten(np.sign(J))
    J = np.abs(np.ndarray.flatten(J))
    np.where(J > cutoff_percentile, J, 0)

    return np.reshape(J * sign_J, (rows, cols))

# makes a dwave bqm and corresponding networkx graph
def make_bqm(h, J, fix_var):
    bqm_nx = nx.from_numpy_matrix(J)
    attr_dict, h_list = {}, []
    for i in range(np.size(h)):
        attr_dict['h_bias'] = h[i]
        h_list.append((i, attr_dict['h_bias']))
    bqm_nx.add_nodes_from(h_list)
    
    bqm = dimod.from_networkx_graph(bqm_nx, vartype='SPIN', node_attribute_name='h_bias', edge_attribute_name='weight')
    fixed_dict = None
    if fix_var:
        lowerE, fixed_dict = dwplb.roof_duality(bqm, strict=True)
        if fixed_dict == {}:
            lowerE, fixed_dict = dwplb.roof_duality(bqm, strict=False)
        for i in fixed_dict.keys():
            bqm.fix_variable(i, fixed_dict[i])

    return bqm, bqm_nx, fixed_dict

# Independent function for quantum error correction
# -> should hold a single correction scheme to start (can always add)
def default_qac(h, J, fix_var, C):
    rows, cols = np.shape(J)
    con_J = J + J.T + np.max(np.ndarray.flatten(J))*np.eye(rows, cols)
    qa_J = np.zeros(C * (rows, cols))
    for i in np.arange(C):
        for j in np.arange(i, C):
            if i==j:
                qa_J[i*rows : (i+1)*rows, j*cols : (j+1)*cols] = J
            else:
                qa_J[i*rows : (i+1)*rows, j*cols : (j+1)*cols] = con_J
    h = np.repeat(h, C) * C

    return make_bqm(h, qa_J, fix_var)

# Used to compare/benchmark the performance of the error correction
def default_copy(h, J, fix_var, C):
    rows, cols = np.shape(J)
    cp_J = np.zeros(C * (rows, cols))
    for i in np.arange(C):
        cp_J[i*rows : (i+1)*rows, i*cols : (i+1)*cols] = J
    h = np.repeat(h, C)

    return make_bqm(h, cp_J, fix_var)

# adjust weights
def scale_weights(th, tJ, strength):
    for k in list(th.keys()):
        th[k] /= strength 
    for k in list(tJ.keys()):
        tJ[k] /= strength

    return th, tJ

# Does the actual DWave annealing and connecting
def dwave_connect(config, iter, sampler, bqm, bqm_nx, anneal_time):
    '''
        TODO: Determine if its possible to reuse the embedding from previous steps
    '''
    num_nodes = bqm.num_variables
    qaresults = np.zeros((config.ngauges[iter]*config.nread, num_nodes))
    for g in range(config.ngauges[iter]):
        a = np.sign(np.random.rand(num_nodes) - 0.5)
        if config.embedding is None:
            config.embedding = minorminer.find_embedding(bqm_nx, sampler.to_networkx_graph())
        th, tJ = dwe.embed_ising(nx.classes.function.get_node_attributes(bqm_nx, 'h_bias'), nx.classes.function.get_edge_attributes(bqm_nx, 'weight'), config.embedding, sampler.adjacency)
        th, tJ = scale_weights(th, tJ, config.strengths[iter])

        qaresult = sampler.sample_ising(th, tJ, num_reads=config.nread, annealing_time=anneal_time, answer_mode='raw')
        unembed_qaresult = dw.embedding.unembed_sampleset(qaresult, config.embedding, bqm)

        for i in range(len(unembed_qaresult.record.sample)):
            unembed_qaresult.record.sample[i, :] = unembed_qaresult.record.sample[i, :] * a
        qaresults[g*config.nread:(g+1)*config.nread] = unembed_qaresult.record.sample

    return qaresults

# Modifies results by undoing variable fixing
def unfix(qaresults, h_len, fixed_dict):
    spins = np.zeros((np.size(qaresults, 0), h_len))
    j = 0
    for i in range(h_len):
        if i in fixed_dict:
            spins[:, i] = fixed_dict[i]
        else:
            spins[:, i] = qaresults[:, j]
            j += 1

    return spins

# Modifies results by undoing variable encoding (qac method)
#   -> uses majority voting
def decode_qac(qaresults, h_len, orig_len, fix_vars, fixed_dict=None):
    if fix_vars:
        spins = unfix(qaresults, h_len, fixed_dict=fixed_dict)

    true_spins = np.zeros(np.size(qaresults, 0), orig_len)
    for i in range(orig_len):
        qubit_copies = []
        for j in range(np.size(spins, 0) / orig_len):
            qubit_copies.append(spins[j*orig_len])
        true_spins[i] = np.sign(stat.mean((qubit_copies)))
    true_spins[np.where(true_spins == 0)] = np.random.choice([-1, 1])

    return true_spins

# Modifies results by undoing variable encoding (copy method)
def decode_copy(qaresults, h_len, orig_len, fix_vars, fixed_dict=None):
    if fix_vars:
        spins = unfix(qaresults, h_len, fixed_dict=fixed_dict)
    
    multi_spins = np.zeros(np.size(spins, 0) / orig_len)       # will be an exact int, not a floating point
    for i in range(np.size(multi_spins)):
        multi_spins[i] = spins[:, i*orig_len : (i+1)*orig_len]

    return multi_spins

# Derives the energies obtained from the annealing
def energies(spins, sigma, qaresults, C_i, C_ij, mu):
    '''
        TODO: check ordering of einsum indecies; maybe add .T
    '''
    en_energies = np.zeros(np.size(qaresults, 0))
    np.sign(spins, spins)

    en_energies = -2 * np.einsum('ij, j', spins, C_i) * sigma
    en_energies += 2 * np.einsum('ij, j', np.einsum('ij, jk', spins, np.triu(C_ij, k=1)), spins[0][:]) * pow(sigma, 2)
    en_energies += 2 * np.einsum('ij, j', np.einsum('ij, jk', spins, C_ij), mu)

    return en_energies

# Derives the energies obtained from the annealing (qac method)
#   -> bc majority vote happens earlier, qac has the same structure as non-qac
def energies_qac(truespins, sigma, qaresults, C_i, C_ij, mu):
    return energies(truespins, sigma, qaresults, C_i, C_ij, mu)

# Derives the energies obtained from the annealing (copy method)
def energies_copy(multi_spins, sigma, qaresults, C_i, C_ij, mu):
    multi_energies = np.zeros(np.size(multi_spins, 0))
    for i in range(np.size(multi_spins, 0)):
        multi_energies[i] = energies(multi_spins[i], sigma, qaresults, C_i, C_ij, mu)

    return multi_energies

# Picks out the unique energies out of the ones derived from energies
def unique_energies(en_energies, energy_fraction, max_state):
    '''
        TODO:
    '''
    unique_energies, unique_indices = np.unique(en_energies, return_index=True)
    ground_energy = np.amin(unique_energies)
    if ground_energy < 0:
        threshold_energy = (1 - energy_fraction) * ground_energy
    else:
        threshold_energy = (1 + energy_fraction) * ground_energy
    lowest = np.where(unique_energies < threshold_energy)
    unique_indices = unique_indices[lowest]
    if np.size(unique_indices, 0) > max_state:
        sorted_indices = np.argsort(en_energies[unique_indices])[-max_state :]
        unique_indices = unique_indices[sorted_indices]
    
    return unique_indices

# Picks out the unique energies out of the ones derived from energies (qac method)
#   -> bc majority vote happens earlier, qac has the same structure as non-qac
def uniques_qac(en_energies, energy_fractions, max_states):
    return unique_energies(en_energies, energy_fractions, max_states)

# Picks out the unique energies out of the ones derived from energies (copy method)
def uniques_copy(multi_energies, energy_fractions, max_states):
    '''
        TODO:
    '''
    multi_uniques = np.zeros(np.size(multi_energies, 0))
    for i in range(np.size(multi_energies, 0)):
        multi_uniques[i] = energies(multi_energies[i], energy_fractions, max_states)

    return multi_uniques

def anneal(config, iter, env, mu):
    ''' Controls the annealing and energy selection process.

        TODO:

        Called from train() multiple times. Each call to anneal performs a training step
        on the D-Wave hardware.

        Parameters:
        - config       Configuration struct (hold many useful config params)
        - iter         Declares the current iteration within the training method.
        - mu           Vector of expected values of qubit spins.
                       This is needed because some configuration parameters are per-iteration arrays.
        - env          The TrainEnv variable that holds the environment (data) used for training.
        - anneal_time  How long the annealing occurs. The longer you anneal, the better the
                       result. But the longer you anneal, the higher the chance for decoherence
                       (which yields garbage results). Units = microseconds.
    '''
    h, J = make_h_J(env.C_i, env.C_ij, mu, pow(config.zoom_factor, iter))
    if config.prune_vars is not None:
        J = config.prune_vars(J, config.cutoff)
    if config.encode_vars is None:
        bqm, bqm_nx, fixed_dict = make_bqm(h, J, config.fix_vars)
        if config.fixed_dict is None:
            config.fixed_dict = fixed_dict
        elif config.fixed_dict.keys() == fixed_dict.keys():
            config.same_fixed_dict_counter += 1
        else:
            config.diff_fixed_dict_counter += 1
    else:
        orig_len = np.size(h)
        bqm, bqm_nx, fixed_dict = config.encode_vars(h, J, config.fix_vars, config.encoding_depth)

    qaresults = dwave_connect(config, iter, env.sampler, bqm, bqm_nx, config.anneal_time)

    if config.encode_vars is None and config.fix_vars is not None:
        spins = unfix(qaresults, np.size(h), fixed_dict=fixed_dict)
    elif config.encode_vars is not None:
        spins = config.decode_vars(qaresults, np.size(h), orig_len, config.fix_vars, fixed_dict=fixed_dict)
        en_energies = config.encoded_energies(spins, pow(config.zoom_factor, iter), qaresults, env.C_i, env.C_ij, mu)
        unique_indices = config.encoded_uniques(en_energies, config.energy_fractions[iter], config.max_states[iter])

        if len(np.shape(spins)) == 2:
            excited_states_arr = [spins[unique_indices]]
        elif len(np.shape(spins)) == 3:
            excited_states_arr = []
            for i in range(np.size(spins, 0)):
                excited_states_arr[i] = spins[i][unique_indices]
        
        return excited_states_arr
    else:
        spins = qaresults

    en_energies = energies(spins, pow(config.zoom_factor, iter), qaresults, env.C_i, env.C_ij, mu)
    unique_indices = unique_energies(en_energies, config.energy_fractions[iter], config.max_states[iter])    
    excited_states_arr = [spins[unique_indices]]
    
    return excited_states_arr


import time

import numpy as np

from dwave.system.samplers import DWaveSampler

# Numpy arrays should be row-major for best performance

class TrainEnv:
    # require for the data to be passed as NP data arrays 
    # -> clear to both user and code what each array is
    # X_train, y_train should be formatted like scikit-learn data arrays are
    # -> X_train is the train data, y_train is the train labels
    def __init__(self, X_train, y_train, endpoint_url, account_token, X_val=None, y_val=None, fidelity=7):
        self.X_train = X_train
        self.y_train = y_train
        if X_val is None:
            self.create_val_data()
        else:
            self.X_val = X_val
            self.y_val = y_val
        
        self.fidelity = fidelity
        self.fidelity_offset = 0.0225 / fidelity

        self.c_i = None
        self.C_i = None
        self.C_ij = None
        self.data_preprocess()

        self.sampler = DWaveSampler(endpoint=endpoint_url, token=account_token, solver=dict(topology__type='pegasus'))

    def create_val_data(self):
        dummy_xt, dummy_xv = np.split(self.X_train, [int(8*np.size(self.X_train, 0)/10)], 0)
        dummy_yt, dummy_yv = np.split(self.y_train, [int(8*np.size(self.y_train)/10)])

        self.X_train, self.X_val = np.array(list(dummy_xt)), np.array(list(dummy_xv))
        self.y_train, self.y_val = np.array(list(dummy_yt)), np.array(list(dummy_yv))

    def data_preprocess(self):
        """
            This duplicates the parameters 'fidelity' times. The purpose is to turn the weak classifiers
            from outputing a single number (-1 or 1) to outputting a binary array ([-1, 1, 1,...]). The 
            use of such a change is to trick the math into allowing more nuance between a weak classifier
            that outputs 0.1 from a weak classifier that outputs 0.9 (the weak classifier outputs are continuous)
            -> thereby discretizing the weak classifier's decision into more pieces than binary.
        """
        """
            This creates a periodic array to shift the outputs of the repeated weak classifier, so that there
            is a meaning to duplicating them. You can think of each successive digit of the resulting weak classifier
            output array as being more specific about what the continuous output was - ie >0, >0.1, >0.2 etc. This 
            description is not exactly correct in this case but it is the same idea as what we're doing.
        """
        m_events, n_params = np.shape(self.X_train) # [M events (rows) x N parameters (columns)]

        c_i = np.repeat(self.X_train, repeats=self.fidelity, axis=1) # [M events (rows) x N*fidelity parameters (columns)]
        
        offset_array = self.fidelity_offset * (np.tile(np.arange(self.fidelity), m_events * n_params) - self.fidelity//2)
        c_i = np.sign(np.ndarray.flatten(c_i, order='C') - offset_array) / (n_params * self.fidelity)
        c_i = np.reshape(c_i, (m_events, n_params*self.fidelity))
        self.c_i = c_i

        C_i = np.einsum('i, ij', self.y_train, c_i)
        C_ij = np.einsum('ji,jk', c_i, c_i)

        self.C_i, self.C_ij = C_i, C_ij


import qamlz

def test_assert():
    assert 2 == 2

def test_qamlz_TrainEnv():
    assert qamlz.TrainEnv == 'Hola, amigo!'

def test_qamlz_ModelConfig():
    assert qamlz.ModelConfig == 'Woah there, muchacho.'

def test_qamlz_Model():
    assert qamlz.Model == 'Woah there, muchacho.'


import datetime

import numpy as np

from sklearn.metrics import accuracy_score

from .anneal import anneal
from .anneal import default_prune, default_qac, decode_qac, energies_qac, uniques_qac

# Used to calculate the total hamiltonian of a certain problem
def total_hamiltonian(mu, sigma, C_i, C_ij):
    ''' Derived from Eq. 9 in QAML-Z paper (ZLokapa et al.)
        Dot products of upper triangle

        TODO: Check indecies; maybe add .T
    '''
    ham = np.sum(-C_i + np.sum(np.einsum('ij, j', np.triu(C_ij, k=1), mu))) * sigma
    ham += np.sum(np.triu(C_ij, k=1)) * pow(sigma, 2)
    
    return ham

# Returns the ML algorithm's predictions
def evaluate(X_data, weights):
    '''
        TODO:
    '''
    return np.sign(np.dot(weights, X_data))

class ModelConfig:
    def __init__(self, n_iterations=10, zoom_factor=0.5):
        self.n_iterations = n_iterations
        self.zoom_factor = zoom_factor
        self.anneal_time = 5
        
        self.flip_probs = np.array([0.16, 0.08, 0.04, 0.02] + [0.01]*(n_iterations - 4))
        self.flip_others_probs = np.array([0.16, 0.08, 0.04, 0.02] + [0.01]*(n_iterations - 4)) / 2

        self.strengths = [3.0, 1.0, 0.5, 0.2] + [0.1]*(n_iterations - 4)
        self.energy_fractions = [0.08, 0.04, 0.02] + [0.01]*(n_iterations - 3)
        self.ngauges = [50, 10] + [1]*(n_iterations - 2)
        self.max_states = [16, 4] + [1]*(n_iterations - 2)
        self.nread = 200

        self.embedding = None
        self.fixed_dict = None
        self.same_fixed_dict_counter = 0
        self.diff_fixed_dict_counter = 0

        self.fix_vars = True
        self.prune_vars = default_prune
        self.cutoff = 95
        self.encode_vars = default_qac
        self.encoding_depth = 3      # from nested qac paper
        self.decode_vars = decode_qac
        self.encoded_energies = energies_qac
        self.encoded_uniques = uniques_qac

class Model:
    '''
        TODO: finish results dict -> add in method for final test accuracy (AUROC) computation
    '''
    def __init__(self, config, env):
        # add in hyperparameters in ModelConfig
        # -> this is where user determines how the model will train
        self.config = config
        self.env = env

        self.start_time = datetime.datetime.now().strftime('%Y/%m/%d__%H-%M-%S')
        self.anneal_results = {}
        self.mus_dict = {}

    def pick_excited_states(self, config, env, iter, excited_states, mu, train_size):
        for excited_state in excited_states:
            new_sigma = pow(config.zoom_factor, iter+1)
            new_mu = mu + (pow(config.zoom_factor, iter) * excited_state)
            new_energy = total_hamiltonian(new_mu, new_sigma, env.C_i, env.C_ij) / (train_size - 1)
            flips = np.ones(np.size(excited_state))
            for qubit in range(np.size(excited_state)):
                temp_s = np.copy(excited_state)
                temp_s[qubit] = 0
                flipped_mu = mu + (pow(config.zoom_factor, iter) * temp_s)
                old_energy = total_hamiltonian(flipped_mu, new_sigma, env.C_i, env.C_ij) / (train_size - 1)
                energy_diff = new_energy - old_energy
                if energy_diff > 0:
                    flip_prob = config.flip_probs[iter]
                    flip = np.random.choice([-1, 1], size=1, p=[1-flip_prob, flip_prob])[0]
                    flips[qubit] = flip
                else:
                    flip_prob = config.flip_others_probs[iter]
                    flip = np.random.choice([-1, 1], size=1, p=[1-flip_prob, flip_prob])[0]
                    flips[qubit] = flip
            flipped_s = excited_state * flips

        return (mu + flipped_s*new_sigma)
            

    def train(self):
        '''
            TODO: recombine weights into actual parameter classifiers
        '''
        mus = [np.ones(np.size(self.env.C_i))]

        for i in range(self.config.n_iterations):
            new_mus = []
            for mu in mus:
                excited_states_arr = anneal(self.config, i, self.env, mu)
                for excited_states in excited_states_arr:
                    new_mus.append([self.pick_excited_states(self.config, self.env, i, excited_states, mu, np.shape(self.env.X_train)[0])])
            accuracies = np.zeros(len(new_mus))
            for j in range(len(new_mus)):
                avg_arr_val =[]
                for mu in new_mus[j]:
                    avg_arr_val.append(accuracy_score(self.env.y_val, evaluate(mu, self.env.c_i)))
                np.append(accuracies, np.mean(np.array(avg_arr_val)))
            mus = new_mus[np.where(np.max(accuracies))]
            mus_filename = 'mus%05d_iter%d_run%d-%s.npy' % (np.shape(self.env.X_train)[0], i, j, self.start_time)
            self.mus_dict[mus_filename] = np.array(mus)




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='qamlz',
    version='0.3.0',
    description='Binary Classifier trained with D-Wave\'s Quantum Annealers.',
    packages=find_packages(include=['qamlz']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "numpy >= 1.20.3",
        "scikit_learn >= 1.0.1",
        "scipy >= 1.7.1",
        "dwave-ocean-sdk >= 4.2.0",
    ],
    extras_require={
        "dev":[
            "pytest >= 3.7",
            "check-manifest >= 0.47",
        ],
    },
    url="https://github.com/tcoulvert/qaml-z",
    author="Thomas Sievert",
    author_email="tcsievert@gmail.com",
)